
from parse import parse
from pathlib import Path
from pydantic import ValidationError
from typing import Union

from . import structure


class DataStructureError(Exception):
    pass


class BaseReader:
    parent_structure = structure.BaseDataStructure

    @staticmethod
    def _read_data_from_file(path: Union[str, Path]):
        with open(path, "r") as f:
            data = f.read()
        return data

    @classmethod
    def load(cls, path: Union[str, Path], parallel: bool = True):
        data = cls._read_data_from_file(path)

        # Loop over all data structure subclasses and attempt to parse data:
        for structure in cls.parent_structure.children:
            meta = structure.extract_meta(data)
            if meta is None:
                continue
            meta = cls.parent_structure.validate_meta(meta)
            table_rows = structure.extract_table(data, meta, parallel)
            if table_rows is not None:
                table_rows = cls.parent_structure.validate_table(table_rows, parallel)
                break

        # Raise DataStructureError if unable to parse:
        if (meta is None) or (table_rows is None):
            raise DataStructureError(
                f"Unable to parse data using an existing data structure definition. You "
                f"may need to add a new data structure definition if one does not exist "
                f"(see 'captif_data_structure.structure')"
            )

        return (meta, table_rows, structure.id)


class DemoReader(BaseReader):
    parent_structure = structure.DemoDataStructure


class DeflectionBeamReader(BaseReader):
    parent_structure = structure.DeflectionBeamDataStructure


class LapCountReader(BaseReader):
    parent_structure = structure.LapCountDataStructure


class TextureReader(BaseReader):
    parent_structure = structure.TextureDataStructure




from datetime import datetime


def tab_to_comma(contents):
    return ",".join(contents.split("\t"))


def combine_date_time_fields(row, delete=False):
    row["datetime"] = datetime.combine(row["date"], row["time"])
    if delete:
        del(row["date"], row["time"])
    return row




from datetime import date, time
from pydantic import BaseModel

from .base import BaseDataStructure


"""
Contains demo data structure definitions.

New data structures must be defined as a new class that derives from the
DemoDataStructure class. These subclasses must be named with a random
8-digit hex (e.g. `uuid4().hex[:8]`) prepended with an underscore ("_").

The subclasses must define:
    1. data_structure class attribute (according to the "parse" package); and
    2. row_model class (Pydantic model) with field order that matches the table column
       order.

data_structure fields must match those in DemoDataStructure.meta_model, which
can be updated with new fields provided it remains backward compatible.

row_model must be compatable with DemoDataStructure.row_model, which is used
ensure consistency between all schema subclasses. Again
DemoDataStructure.row_model can be updated with new fields provided it remains backward compatible.

"""


class DemoDataStructure(BaseDataStructure):
    """
    Demo data structure parent class.

    All demo data structure classes must be derived from this parent class.
    data_structure and row_model should be redefined in the subclasses.

    """

    class meta_model(BaseModel):
        """
        Meta data Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        param_1: str
        param_2: float

    class row_model(BaseModel):
        """
        Table row Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        column_1: str
        column_2: int


class _abcd1234(DemoDataStructure):

    data_structure = (
        "param1\t{param_1}\n"
        "param2\t{param_2:f}\n"
        "param3\t{}\n"
        "\n"
        "column B\tcolumn A\n"
        "{}"
    )

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        column_1: str
        column_2: int

    @staticmethod
    def meta_preprocessor(meta: dict):
        """
        Meta data preprocessor.
        """
        meta["param_2"] += 10
        return meta

    @staticmethod
    def row_preprocessor(row: dict):
        """
        Table row preprocessor.
        """
        row["column_1"] = row["column_1"] + "_"
        row["column_2"] = row["column_2"] - 1
        return row




import psutil
import os
import numpy as np
from typing import List, Optional
from parse import parse
from platform import system
from pydantic import BaseModel, ValidationError
from unsync import unsync
from multiprocessing import cpu_count

from ..helpers import tab_to_comma


CPU_COUNT = cpu_count()
OS = system()


class BaseDataStructure:
    """
    Base data structure class
    """

    data_structure = ""  # Data structure string (as per the "parse" package)

    class meta_model(BaseModel):
        """
        Meta data Pydantic model.
        """
        pass

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        pass

    @staticmethod
    def meta_preprocessor(meta: dict) -> dict:
        """
        Meta data preprocessor.

        This method is used to convert the meta data from a given data structure subclass
        in to the meta data format of the parent data structure. A common use case is
        combining "date" and "time" meta data fields into a single "datetime" field.
        """
        return meta

    @staticmethod
    def row_preprocessor(row: dict) -> dict:
        """
        Table row preprocessor.

        This method is used to convert a table row from a given data structure subclass
        in to the table row format of the parent data structure. A common use case is
        combining "date" and "time" meta data fields into a single "datetime" field, 
        converting units or deleting unused columns.

        For more complex preprocessing requiring knowledge of all rows and meta data use
        the `table_preprocessor` method.
        """
        return row

    @staticmethod
    def table_preprocessor(table_rows: List[dict], meta: dict = {}) -> List[dict]:
        """
        Table preprocessor.

        This method is used as a final step before the table rows are validated against
        the table row format of the parent data structure. It is called after the
        `row_preprocessor` method and performs a similar function execpt that the method
        has access to all table rows and any meta data. A common use case is to add an
        sample spacing column, which requires knowledge of both the sample rate (which
        can be passed in as meta data) and the total number of readings.
        """
        return table_rows

    @classmethod
    def extract_meta(cls, data: str) -> dict:
        """
        Extract the meta data from the data sting.

        The meta data is run through meta_preprocessor before being returned.
        """
        parsed = parse(cls.data_structure, data)
        if parsed is None:
            return parsed

        # Run the meta data through the meta_preprocessor and return:
        return cls.meta_preprocessor(parsed.named)

    @classmethod
    def extract_table(
        cls, data: str, meta: Optional[dict] = None, parallel: bool = True,
    ) -> dict:
        """
        Extract the table rows from the data string.

        The table rows are run through the row_preprocessor before being returned.
        """
        parsed = parse(cls.data_structure, data)
        if parsed is None:
            return parsed

        # Extract table data from the last unnamed parameter in data_structure and
        # convert to list of dicts using the row_model fields as the keys:
        table_data = parsed.fixed[-1]
        table_data = [row.split(",") for row in tab_to_comma(table_data).split("\n")]
        table_rows = [dict(zip(cls.row_model.__fields__, row)) for row in table_data]

        # Attempt to validate table rows using the row_model. This will attempt to cast
        # any fields that don't match the corresponding row_model field.
        try:
            table_rows = validate_table_rows(cls.row_model, table_rows, parallel=parallel)
        except ValidationError:
            return None  # return None if unable to validate table rows

        # Run the table rows through row_preprocessor:
        table_rows = preprocess_table_rows(cls.row_preprocessor, table_rows, parallel)

        # Run the table rows through table_preprocessor and return:
        return cls.table_preprocessor(table_rows, meta)

    @classmethod
    def validate_meta(cls, meta):
        """
        Validate the meta data using meta_model.
        """
        return cls.meta_model(**meta).dict(exclude_unset=True)

    @classmethod
    def validate_table(cls, table_rows, parallel=True):
        """
        Validate the table rows using row_model.
        """
        return validate_table_rows(
            cls.row_model, table_rows, exclude_unset=True, parallel=parallel,
        )

    @classmethod
    @property
    def children(cls) -> list:
        """
        List of data structure classes derived from this parent class.
        """
        return cls.__subclasses__()

    @classmethod
    @property
    def id(cls) -> str:
        """
        Data structure ID.
        """
        return cls.__name__.split("_")[-1]


def unpack_results(tasks):
    return [rr for tt in tasks for rr in tt.result()]


def limit_cpu():
    p = psutil.Process(os.getpid())
    if OS == "Windows":
        p.nice(psutil.BELOW_NORMAL_PRIORITY_CLASS)
    else:
        p.nice(19)


@unsync(cpu_bound=True)
def call_row_model(row_model, rows, exclude_unset=False):
    limit_cpu()
    return [row_model(**row).dict(exclude_unset=exclude_unset) for row in rows]


@unsync(cpu_bound=True)
def call_row_preprocessor(row_preprocessor, rows):
    limit_cpu()
    return [row_preprocessor(row) for row in rows]


def validate_table_rows(row_model, table_rows, exclude_unset=False, parallel=True):
    if parallel:
        tasks = [
            call_row_model(row_model, rows.tolist(), exclude_unset)
            for rows in np.array_split(table_rows, CPU_COUNT)
        ]
        return unpack_results(tasks)
    return [row_model(**row).dict(exclude_unset=exclude_unset) for row in table_rows]


def preprocess_table_rows(row_preprocessor, table_rows, parallel=True):
    if parallel:
        tasks = [
            call_row_preprocessor(row_preprocessor, rows.tolist())
            for rows in np.array_split(table_rows, CPU_COUNT)
        ]
        return unpack_results(tasks)
    return [row_preprocessor(row) for row in table_rows]




from datetime import date, datetime, time
from pydantic import BaseModel

from .base import BaseDataStructure


"""
Contains data structure definitions for deflection beam data.

New data structures must be defined as a new class that derives from the
DeflectionBeamDataStructure class. These subclasses must be named with a random
8-digit hex (e.g. `uuid4().hex[:8]`) prepended with an underscore ("_").

The subclasses must define:
    1. data_structure class attribute (according to the "parse" package); and
    2. row_model class (Pydantic model) with field order that matches the table column
       order.

data_structure fields must match those in DeflectionBeamDataStructure.meta_model, which
can be updated with new fields provided it remains backward compatible.

row_model must be compatable with DeflectionBeamDataStructure.row_model, which is used
ensure consistency between all schema subclasses. Again
DeflectionBeamDataStructure.row_model can be updated with new fields provided it remains backward compatible.

"""


class DeflectionBeamDataStructure(BaseDataStructure):
    """
    Deflection beam data structure parent class.

    All deflection beam data structure classes must be derived from this parent class.
    data_structure and row_model should be redefined in the subclasses.

    """

    class meta_model(BaseModel):
        """
        Meta data Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        datetime: datetime
        station_no: int
        max_raw_deflection_mm: float

    class row_model(BaseModel):
        """
        Table row Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        distance_m: float
        raw_deflection_mm: float


class _cb379c14(DeflectionBeamDataStructure):

    data_structure = (
        "Date:\t{date:tg}\n"
        "Time:\t{time:tt}\n"
        "Location:\t{station_no:d}\n"
        "Operator:\t{}\n"
        "Latitude\t{}\n"
        "Longitude\t{}\n"
        "Max Deflection Measured\t{max_raw_deflection_mm:f}\n"
        "Max Deflection Corrected\t{}\n"
        "Bowl Corrected\t{}\n"
        "CF [Dpk-D200]\t{}\n"
        "Distance (m)\tDeflection  (mm)\n"
        "{}"
    )

    @staticmethod
    def meta_preprocessor(meta):
        """
        Meta data preprocessor.
        """
        meta["datetime"] = datetime.combine(meta["date"].date(), meta["time"])
        del(meta["date"], meta["time"])
        return meta

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        distance_m: float
        raw_deflection_mm: float




from datetime import date, datetime, time
from typing import Optional
from pydantic import BaseModel, validator

from .base import BaseDataStructure
from ..helpers import combine_date_time_fields


"""
Contains data structure definitions for CAPTIF lap count data.

New data structures must be defined as a new class that derives from the
LapCountDataStructure class. These subclasses must be named with a random
8-digit hex (e.g. `uuid4().hex[:8]`) prepended with an underscore ("_").

The subclasses must define:
    1. data_structure class attribute (according to the "parse" package); and
    2. row_model class (Pydantic model) with field order that matches the table column
       order.

data_structure fields must match those in LapCountDataStructure.meta_model, which
can be updated with new fields provided it remains backward compatible.

row_model must be compatable with LapCountDataStructure.row_model, which is used
ensure consistency between all schema subclasses. Again
LapCountDataStructure.row_model can be updated with new fields provided it remains backward compatible.

"""


class LapCountDataStructure(BaseDataStructure):
    """
    CAPTIF lap count data structure parent class.

    All lap count data structure classes must be derived from this parent class.
    data_structure and row_model should be redefined in the subclasses.

    """

    class meta_model(BaseModel):
        """
        Meta data Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        pass

    class row_model(BaseModel):
        """
        Table row Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        lap_count: int
        datetime: datetime
        position_cm: Optional[int]
        speed_kph: Optional[int]


class _d26612dd(LapCountDataStructure):

    data_structure = (
        "lap_count\tdate\ttime\n"
        "{}"
        "\n"
    )

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        lap_count: int
        date: date
        time: time

        @validator("date", pre=True)
        def parse_date(cls, value):
            return datetime.strptime(
                value,
                "%d/%m/%Y"
            ).date()

        @validator("time", pre=True)
        def parse_time(cls, value):
            return datetime.strptime(
                value,
                "%I:%M %p"
            ).time()

    @staticmethod
    def row_preprocessor(row: dict):
        """
        Table row preprocessor.
        """
        return combine_date_time_fields(row, delete=True)


class _e7b769c1(LapCountDataStructure):

    data_structure = (
        "lap_count\tdate\ttime\tposition_cm\tspeed_kph\n"
        "{}"
        "\n"
    )

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        lap_count: int
        date: date
        time: time
        position_cm: int
        speed_kph: int

        @validator("date", pre=True)
        def parse_date(cls, value):
            return datetime.strptime(
                value,
                "%d/%m/%Y"
            ).date()

        @validator("time", pre=True)
        def parse_time(cls, value):
            return datetime.strptime(
                value,
                "%I:%M:%S %p"
            ).time()

    @staticmethod
    def row_preprocessor(row: dict):
        """
        Table row preprocessor.
        """
        return combine_date_time_fields(row, delete=True)



from datetime import date, datetime, time
from typing import Any, List, Optional
from pydantic import BaseModel, validator

from .base import BaseDataStructure
from ..helpers import combine_date_time_fields


"""
Contains data structure definitions for CAPTIF texture data.

New data structures must be defined as a new class that derives from the
TextureDataStructure class. These subclasses must be named with a random
8-digit hex (e.g. `uuid4().hex[:8]`) prepended with an underscore ("_").

The subclasses must define:
    1. data_structure class attribute (according to the "parse" package); and
    2. row_model class (Pydantic model) with field order that matches the table column
       order.

data_structure fields must match those in TextureDataStructure.meta_model, which
can be updated with new fields provided it remains backward compatible.

row_model must be compatable with TextureDataStructure.row_model, which is used to
ensure consistency between all schema subclasses. Again TextureDataStructure.row_model
can be updated with new fields provided it remains backward compatible.

"""


class TextureDataStructure(BaseDataStructure):
    """
    CAPTIF texture data structure parent class.

    All texture data structure classes must be derived from this parent class.
    data_structure and row_model should be redefined in the subclasses.

    """

    class meta_model(BaseModel):
        """
        Meta data Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        date: Optional[date]
        file_number: Optional[int]
        sample_spacing_mm: Optional[float]

        @validator("date", pre=True)
        def parse_date(cls, value):
            return datetime.strptime(
                value,
                "%d/%m/%Y"
            ).date()

    class row_model(BaseModel):
        """
        Table row Pydantic model.

        Must be backward compatable i.e. do not modify the data types and either make any
        new parameters optional or specify default values.

        """
        distance_mm: float
        relative_height_mm: Optional[float]


class _4102a5dd(TextureDataStructure):

    data_structure = (
        "Road Name\tCAPTIF\n"
        "Ref Station\t{}\n"
        "Start Pos (m)\t{}\n"
        "End Pos (m)\t{}\n"
        "Direction\t{}\n"
        "Wheel Path\tLeft\n"
        "Date\t{date}\n"
        "File No.\t{file_number}\n"
        "Current Pos\t{}\n"
        "*****DATA*****\t\n"
        "{}"
        "\n"
    )

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        point_no: Any
        distance_mm: float
        relative_height_mm: Optional[float]

        @validator("relative_height_mm", pre=True)
        def parse_relative_height_mm(cls, value):
            return None if value == "NaN" else value

    @staticmethod
    def row_preprocessor(row: dict) -> dict:
        del(row["point_no"])
        return row


class _245ff223(TextureDataStructure):

    data_structure = (
        "Road Name\tCAPTIF\t\n"
        "Ref Station\t{}\t\n"
        "Start Pos (m)\t{}\t\n"
        "End Pos (m)\t{}\t\n"
        "Direction\t{}\t\n"
        "Wheel Path\tLeft\t\n"
        "Date\t{date}\t\n"
        "File No.\t{file_number}\t\n"
        "Current Pos\t{}\t\n"
        "*****DATA*****\t\t\n"
        "Data Point No:\tDistance (mm)\tDepth (mm)\n"
        "{}"
        "\n"
    )

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        point_no: Any
        distance_mm: float
        relative_height_mm: Optional[float]

        @validator("relative_height_mm", pre=True)
        def parse_relative_height_mm(cls, value):
            return None if value == "NaN" else value


class _7cd12dee(TextureDataStructure):
    data_structure = (
        "profile_name: {}\n"
        "sample_spacing_mm: {sample_spacing_mm}\n"
        "{}"
        "\n"
    )

    class row_model(BaseModel):
        """
        Table row Pydantic model.
        """
        relative_height_mm: Optional[float]

        @validator("relative_height_mm", pre=True)
        def parse_relative_height_mm(cls, value):
            return None if value == "NaN" else value

    @staticmethod
    def table_preprocessor(table_rows: List[dict], meta: dict = {}) -> List[dict]:
        for ii, _ in enumerate(table_rows):
            table_rows[ii]["distance_mm"] = ii * meta.get("sample_spacing_mm")
        return table_rows



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['captif_data_structures', 'captif_data_structures.structure']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.3.3,<2.0.0',
 'parse>=1.19.0,<2.0.0',
 'psutil>=5.8.0,<6.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'unsync>=1.4.0,<2.0.0']

setup_kwargs = {
    'name': 'captif-data-structures',
    'version': '0.7',
    'description': '',
    'long_description': '# captif-data-structures\n\nPython package to keep track of the various data files and file versions used by CAPTIF Road Research. The package includes reader classes for general data types.\n\nThe following data types are currently supported.\n\nCAPTIF facility:\n- lap count\n\nInstruments:\n - deflection beam\n - stationary laser profilometer (SLP)\n\nExamples of the data structures currently supported can be found in `./tests/data`.',
    'author': 'John Bull',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/captif-nz/captif-data-structures',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)




from .base import BaseDataStructure
from .demo import DemoDataStructure
from .deflection_beam import DeflectionBeamDataStructure
from .lap_count import LapCountDataStructure
from .texture import TextureDataStructure
